/*
 * $Id: Connector.java,v 1.9 2005/12/15 13:59:02 windowsnt\martin Exp $
 * 
 * Copyright (C) ITC 2005 - International Institute for Geo-Information Science and Earth Observation
 *    
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package org.n52.connectors.core;

import org.geotools.metadata.iso.MetadataEntity;
import org.n52.connectors.core.exceptions.IllegalConnectorRequestException;
import org.n52.connectors.core.exceptions.N52Exception;
import org.n52.connectors.core.exceptions.ParameterDoesNotExistException;
import org.w3c.dom.Document;

/**
 * A connector is a support class that is needed to contain all administrative
 * information that is needed to access service providers. It is able to
 * initialize and retrieve/create objects that give access to a service provider
 * To understand the role of the connector, below is a short scenario how a normal request works out
 * <ol>
 * <li>A client formulates a ConnectorRequest by stating that he wants to do an operation, say resample operation using bicubic.</li> 
 * <li>The ConnectorFactory receives the request and passes it the ConnectorRegistry</li>
 * <li>The registry looks if it knows a connector which is able to support this operation. If so it returns to the factory an XML string which contains all information to create such a connector</li>
 * <li>The ConnectorFactory receives the string and through the ClassLoader creates a connector.</li>
 * <li>The newly created connector creates and initializes a ServiceAdapter that gives access to a resource that can do a resample operation (e.g. a library)</li>
 * <li>The object is returned to the client</li>
 * <li>The user formulates all the parameters that are needed for the resampling and passes them the service adapter</li>
 * <li>The parameters are passed to the execute method of the service adapter and it returns one or more objects as result of the operation; or it throws an exception I cases where some erroneous condition exists</li>
 * </ol>
 * @author martin schouwenburg
 */
public interface Connector {

	/**
	 * @param s
	 *            a metadata identifier based on ISO 19115. see..
	 * @return returns a metadataentity based on geotools (ISO 19115)
	 */
	MetadataEntity getMetadata(String s);

	/**
	 * This method checks according to the information in the ServiceRequest if this connector has access to desired functionality. In many cases this is simply a check against a string that is an unique identification for the operation e.g. For the unique identification (see also the paragraph Function names); for resample this would  be : ImageGeometryModel.Conversion.
	 *Operation may or may not have a second tag called method. A method further refines an operation to a specific algorithm. Conceptual the same thing is done but using a different method. In our example the method would be Bicubic. 
	 *These kind of unique identifications are tagged as service in the request. So in code this could look like <br>
	 *<br>
	 *<code>
	 *if ( String(ImageGeometryModel.Conversion).equals( (String)request.get(service)) {<br>
	 *	if ( (String)request.get(method).equals(bicubic)) <br>
	 * 		return true;<br>
	 * }
	 * </code>
	 * <br>
     *This is the only thing it is supported function is supposed to do.
     *When creating a connector one can pick unique ids form the existing list or create new ones if they dont exist yet. The basis of the naming scheme is given by ISO 19119 in their limited taxonomy; in almost all cases one needs to extend on this as it is fairly limited.
	 * 
	 * @param request
	 *            contains all information that is needed to find a suitable
	 *            connector
	 * @return true if the information in the request matches with the
	 *         properties of this connector
	 */
	boolean isSupported(ServiceRequest request);

	/**
	 * Initializes the connector with information contained in the request. This
	 * initialization happens at the client side (factory) with parameters that
	 * do not need to be send to the registry. The method also ensures that
	 * initialization information send by the registry to the factory is
	 * properly used to initialize the connector This methode can be overruled
	 * in an appropraite connector
	 * 
	 * @param xmlConfiguration
	 *            xml DOM document containing the information returned by the
	 *            registry
	 * @param request
	 *            The original request may contain initialization information
	 *            used as the client side
	 * @throws IllegalConnectorRequestException
	 *             requested parameters where not present or invalid in the
	 *             request
	 * @throws ParameterDoesNotExistException
	 */
	void factorySideInitialization(Document xmlConfiguration,
			ServiceRequest request) throws N52Exception;

	/**
	 * Creates an Xml string that contains initialization inforamtion needed by
	 * the factory to properly created the connector implied by the request.
	 * This initialization happens at the server side (registry) with parameters
	 * that are only present at the server side. This methode can be overruled
	 * in an appropraite connector
	 * 
	 * @param request
	 *            The original request may contain information that is needed by
	 *            the server side to create extra initialization information
	 * @throws IllegalConnectorRequestException
	 *             requested parameters where not present or invalid in the
	 *             request
	 */
	String registrySideInitialization(ServiceRequest request);

	/**
	 * access to the name field of a connector
	 * 
	 * @return a name of the connector
	 */
	String getName();

	/**
	 * Returns an object that supports the desired functionality or gives
	 * (indirectly)access to the desired functionality.
	 * 
	 * @param req
	 * @return
	 * @throws N52Exception
	 */
	Object getImplementingObject(ServiceRequest req) throws N52Exception;
}
