package org.n52.connectors.core;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.StringTokenizer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Level;
import org.n52.connectors.core.exceptions.IllegalConnectorRequestException;
import org.n52.connectors.core.exceptions.IllegalDefaultSettingException;
import org.n52.connectors.core.exceptions.ParameterDoesNotExistException;
import org.n52.connectors.core.exceptions.WrappedException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.sun.org.apache.xpath.internal.XPathAPI;

/**
 * Handling of requests on the same system as the factory is located. The local
 * registry is created by a client program and has access to locally defined
 * connectors. These connectors must be found on a path given at creation time
 * of the connector. Note that the namespace of the conenctors is of
 * significance here (that is dont include them in the path).
 * 
 * @author martin schouwenburg
 */
public class LocalConnectorTypeRegistry extends AbstractConnectorTypeRegistry {

	public LocalConnectorTypeRegistry(String path) throws WrappedException,
			IllegalDefaultSettingException {
		connectorPath = path;
		if (path.equals(""))
			connectorPath = "../ConnectorCore";

		try {
			connectorPath = new File(connectorPath).getCanonicalPath();
		} catch (IOException e) {
			// Ignore this error and leave connectorPath as it is
		}

		connectorPath = normalizePath(connectorPath);

		// now connectorPath is canonical and ends with a '/' or '\' (platform
		// dependent)

		LoadDefaultMapping();
		connectorList.clear();
		ReadConnectors(connectorPath, "*.class");

	}

	/**
	 * Loads the defualt mapping for the use of functionality from the
	 * connectorConfiguration.xml and puts it in internal hashmap
	 * 
	 * @throws WrappedException
	 *             contains java exceptions encountered during execution of this
	 *             method.
	 * @throws IllegalDefaultSettingException
	 */
	private void LoadDefaultMapping() throws WrappedException,
			IllegalDefaultSettingException {
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory
					.newInstance();
			DocumentBuilder build;
			build = factory.newDocumentBuilder();
			Document doc = build.parse(connectorPath
					+ "connectorConfiguration.xml");
			NodeList nodes = XPathAPI.selectNodeList(doc.getDocumentElement(),
					"*/ServiceMapping");
			for (int i = 0; i < nodes.getLength(); ++i) {
				Node n = nodes.item(i);
				Node n2 = XPathAPI.selectSingleNode(n, "Service/text()");
				String service = n2.getNodeValue();
				n2 = XPathAPI.selectSingleNode(n, "Default/text()");
				String defaultMapping = n2.getNodeValue();
				setDefaultConnector(service, defaultMapping);
				logger.info("loading default service mapping : service ="
						+ service + " connector=" + defaultMapping);
			}

		} catch (FileNotFoundException e) {
			// pass, if there is no config file. first come first served
			// mechanism will be used to select a connector
		} catch (TransformerException e) {
			throw new WrappedException(e, logger, Level.ERROR);
		} catch (SAXException e) {
			throw new WrappedException(e, logger, Level.ERROR);
		} catch (IOException e) {
			throw new WrappedException(e, logger, Level.ERROR);
		} catch (ParserConfigurationException e) {
			throw new WrappedException(e, logger, Level.ERROR);
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.n52.connectors.core.ConnectorTypeRegistry#getConnectorClass(org.n52.connectors.core.ConnectorRequest)
	 */
	public String getConnectorClass(ServiceRequest request)
			throws IllegalConnectorRequestException, WrappedException,
			ParameterDoesNotExistException {

		try {
			String connectorName = (String) request.get("connector");
			if (connectorName.equals("default") && (defaultMapping != null)) {
				String serviceId = (String) request.get("service");
				connectorName = defaultMapping.get(serviceId.toLowerCase());
			}

			int i = 0;
			// try to find the default that matches
			for (Connector conn : connectorList) {
				String name = conn.getName();
				if (name.equalsIgnoreCase(connectorName)) { // condition 1: the
					// name is
					// right
					String res = getConnectorInfo(request, conn);
					if (res != null)
						return res;
				}
				++i;
			}

			// first attempt failed, second attempt no deafults now
			for (Connector conn : connectorList) {
				String res = getConnectorInfo(request, conn);
				if (res != null)
					return res;

				++i;
			}
		} catch (MalformedURLException e) {
			throw new WrappedException(e, logger, Level.FATAL);
		}
		throw new IllegalConnectorRequestException("Request for "
				+ request.get("service") + " is not supported", logger,
				Level.ERROR);
	}

	/**
	 * @param request
	 * @param conn
	 * @throws MalformedURLException
	 */
	private String getConnectorInfo(ServiceRequest request, Connector conn)
			throws MalformedURLException {
		if (conn.isSupported(request)) { // condition 2:
			// isSupported
			File classFile = new File(connectorPath
					+ conn.getClass().getSimpleName() + ".class");
			URL classUrl = classFile.toURL();
			logger.log(Level.INFO, "returned respons :" + classUrl.toString()
					+ " class :" + conn.getClass().getName());
			return nameToXml(conn, classUrl, request);
		}

		return null;
	}

	/**
	 * Loads all connectors found in connectorFolder into the connectorList
	 * array.
	 * 
	 * @param connectorFolder
	 *            the folder where all connectors reside
	 * @param pattern
	 *            the extension of the connector files (normally ".class")
	 * @throws WrappedException
	 */
	private void ReadConnectors(String connectorFolder, String pattern)
			throws WrappedException {

		class FilePatternFilter implements FilenameFilter {
			public FilePatternFilter(String pat) {
				// pat is ';' delimited list of extensions
				// example: pat = "*.class;*.jar"
				// the prefix can not be fixed (i.e. pat="conn*.class" will not
				// work)
				pattern = "";
				StringTokenizer st = new StringTokenizer(pat, "*"); // filter
				// out *'s :
				// we weed
				// .ext
				// instead
				// of *.ext
				while (st.hasMoreTokens())
					pattern += st.nextToken();
			}

			public boolean accept(File dir, String name) {
				if ((new File(dir, name)).isDirectory())
					return true;
				else {
					String strToBeMatched = name.toLowerCase();
					boolean matchPattern = false;
					StringTokenizer st = new StringTokenizer(pattern, ";");
					while (!matchPattern && st.hasMoreTokens())
						matchPattern = strToBeMatched.endsWith(st.nextToken());
					return matchPattern;
				}
			}

			private String pattern;
		}

		File folder = new File(connectorFolder);
		FilePatternFilter fpf = new FilePatternFilter(pattern);

		File[] connectorFileList = folder.listFiles(fpf);
		for (int i = 0; i < connectorFileList.length; ++i) {
			if (connectorFileList[i].isDirectory()) {
				ReadConnectors(connectorFileList[i].getAbsolutePath(), pattern);
			} else {
				String fileExt = connectorFileList[i].getName();
				int iExtIndex = fileExt.lastIndexOf('.');
				String className;
				if (iExtIndex >= 0)
					className = fileExt.substring(0, iExtIndex);
				else
					className = fileExt;

				addConnector(className, connectorFolder);
			}
		}
	}

	/**
	 * Adds the specified connector to the connectorList member
	 * 
	 * @param className:
	 *            the plain class name of the connector (thus excluding
	 *            namespace)
	 * @param connectorFolder:
	 *            the path where the connector resides: used to generate the
	 *            namespace
	 * @throws WrappedException
	 */
	private void addConnector(String className, String connectorFolder)
			throws WrappedException {
		String nameSpace = normalizePath(connectorFolder.substring(
				connectorPath.length(), connectorFolder.length()));
		nameSpace = nameSpace.replace('/', '.'); // substiute both forward
		// and backward slash for platform independency
		nameSpace = nameSpace.replace('\\', '.');
		className = nameSpace + className;

		try {
			URL classPathUrl = new File(connectorPath).toURL();
			URL[] urlsToLoadFrom = new URL[] { classPathUrl };
			URLClassLoader loader = new URLClassLoader(urlsToLoadFrom);
			Class connectorClass = Class.forName(className, true, loader);
			Connector conn = (Connector) connectorClass.newInstance();
			connectorList.add(conn);
			logger.info("Adding connector to registry :" + conn.getName());
		} catch (InstantiationException e) {
			// ignore; just skip adding this class to the connectorList
		} catch (NoClassDefFoundError e) {
			int k = 0;
			// ignore; just skip adding this class to the connectorList
		} catch (ClassCastException e) {
			// ignore; just skip adding this class to the connectorList
		} catch (IllegalAccessException e) {
			throw new WrappedException(e, logger, Level.FATAL);
		} catch (ClassNotFoundException e) {
			throw new WrappedException(e, logger, Level.FATAL);
		} catch (MalformedURLException e) {
			throw new WrappedException(e, logger, Level.FATAL);
		}
	}

	/**
	 * Composes an XML String that describes a connector.
	 * 
	 * @param connectorName
	 *            the name of the connector
	 * @param classUrl
	 *            the URL location of the connector's class file
	 * @return the two parameters in an XML String
	 */
	private String nameToXml(Connector conn, URL connectorUrl,
			ServiceRequest request) {
		String returnXml = new String(
				"<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		returnXml += "<ConnectorResponse>";
		returnXml += "<classname>" + conn.getClass().getName() + "</classname>";
		returnXml += "<classUrl>" + connectorUrl + "</classUrl>";
		returnXml += conn.registrySideInitialization(request);
		returnXml += "</ConnectorResponse>";
		return returnXml;
	}

	private String normalizePath(String path) {
		if (!path.endsWith("/") && !path.endsWith("\\"))
			path += File.separatorChar;
		return path;
	}

	private ArrayList<Connector> connectorList = new ArrayList<Connector>();

	private String connectorPath;
}
