/* Implementation for class org_n52_connectors_impl_GdalConnector */

#include "org_n52_connectors_impl_GdalConnector.h"

#include <gdal_priv.h> // all GDAL stuff is in this file; make sure to link to gdal_i.lib later on

static bool gdalInitialized = false; // openGdal or createGdal call GDALAllRegister() and set this to true

/*
 * Class:     org_n52_connectors_impl_GdalConnector
 * Method:    getDataBlock
 * Signature: (Lorg/opengis/coverage/grid/GridRange;[D)[D
 */
JNIEXPORT jdoubleArray JNICALL Java_org_n52_connectors_impl_GdalConnector_getDataBlock
  (JNIEnv * env, jobject obj,
		jint xStart, jint yStart, jint xSize, jint ySize,
		jint nBands, jdoubleArray destination)
{
	if (!destination || (env->GetArrayLength(destination) < (xSize * ySize * nBands)))
		destination = env->NewDoubleArray(xSize * ySize * nBands);

	// obtain our GDALDataSet from the java member
	jclass objectCls = env->FindClass("org/n52/connectors/impl/GdalConnector");
	jfieldID fid = env->GetFieldID(objectCls, "gdalDataSet", "J"); // "J" is type "long"
	GDALDataset* gdalDataSet = (GDALDataset*)env->GetLongField(obj, fid);

	jdouble * dataArray = env->GetDoubleArrayElements(destination, 0);
	jdouble * data = dataArray; // additional ptr for looping the bands

	for (int band = 1; band <= nBands; ++band) // GDAL bands start at 1
	{
		GDALRasterBand* gdalRasterBand = gdalDataSet->GetRasterBand(band);
		CPLErr err = gdalRasterBand->RasterIO(GF_Read, xStart, yStart, xSize, ySize, data, xSize, ySize, GDT_Float64, 0, 0); 
		if (CE_Failure == err)
		{
			jclass rangeExceptionCls = env->FindClass("org/opengis/coverage/grid/InvalidRangeException");
			if (rangeExceptionCls == 0) { /* Unable to find the exception class, give up. */
					return NULL; // get rid of warning "not all control paths return a value"
			}
			env->ThrowNew(rangeExceptionCls, "thrown from C code");
			return NULL; // get rid of warning "not all control paths return a value"
		}
		data += xSize * ySize * sizeof(jdouble); // point "data" to next band
	}
	env->ReleaseDoubleArrayElements(destination, dataArray, 0);
	return destination;
}

/*
 * Class:     org_n52_connectors_impl_GdalConnector
 * Method:    setDataBlock
 * Signature: (Lorg/opengis/coverage/grid/GridRange;[D)V
 */
JNIEXPORT void JNICALL Java_org_n52_connectors_impl_GdalConnector_setDataBlock
  (JNIEnv * env, jobject obj, jint xStart, jint yStart, jint xSize, jint ySize, jint nBands, jdoubleArray values)
{
	if (!values || (env->GetArrayLength(values) < (xSize * ySize * nBands)))
	{
		jclass arrayExceptionCls = env->FindClass("java/lang/ArrayIndexOutOfBoundsException");
		if (arrayExceptionCls != 0) /* The exception class was found. */
			env->ThrowNew(arrayExceptionCls, "thrown from C code");
	}

	// obtain our GDALDataSet from the java member
	jclass objectCls = env->FindClass("org/n52/connectors/impl/GdalConnector");
	jfieldID fid = env->GetFieldID(objectCls, "gdalDataSet", "J"); // "J" is type "long"
	GDALDataset* gdalDataSet = (GDALDataset*)env->GetLongField(obj, fid);

	if (gdalDataSet->GetAccess() != GA_Update)
	{
		jclass readonlyExceptionCls = env->FindClass("org/opengis/coverage/grid/GridNotEditableException");
		if (readonlyExceptionCls != 0) /* The exception class was found. */
			env->ThrowNew(readonlyExceptionCls, "thrown from C code");
	}

	jdouble * dataArray = env->GetDoubleArrayElements(values, 0);
	jdouble * data = dataArray; // additional ptr for looping the bands in the (possibly) 3-dimensional dataArray

	for (int band = 1; band <= nBands; ++band) // GDAL bands start at 1
	{
		GDALRasterBand* gdalRasterBand = gdalDataSet->GetRasterBand(band);
		CPLErr err = gdalRasterBand->RasterIO(GF_Write, xStart, yStart, xSize, ySize, data, xSize, ySize, GDT_Float64, 0, 0);
		if (CE_Failure == err)
		{
			jclass rangeExceptionCls = env->FindClass("org/opengis/coverage/grid/InvalidRangeException");
			if (rangeExceptionCls != 0) /* The exception class was found. */
				env->ThrowNew(rangeExceptionCls, "thrown from C code");
		}
		data += xSize * ySize * sizeof(jdouble); // point "data" to next band
	}
	env->ReleaseDoubleArrayElements(values, dataArray, 0);
}

/*
 * Class:     org_n52_connectors_impl_GdalConnector
 * Method:    initializeGdal
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_n52_connectors_impl_GdalConnector_openGdal
  (JNIEnv * env, jobject obj, jstring javaFileName, jboolean readWrite)
{
	if (!gdalInitialized)
	{
		GDALAllRegister();
		gdalInitialized = true;
	}

	const char *strFileName = env->GetStringUTFChars(javaFileName, 0);

	GDALDataset* gdalDataSet = 0;
	if (readWrite)
		gdalDataSet = (GDALDataset*)GDALOpen(strFileName, GA_Update);
	else
		gdalDataSet = (GDALDataset*)GDALOpen(strFileName, GA_ReadOnly);

	env->ReleaseStringUTFChars(javaFileName, strFileName);

	if (0 == gdalDataSet)
	{
			jclass connectorExceptionCls = env->FindClass("org/n52/connectors/core/exceptions/IllegalConnectorRequest");
			if (connectorExceptionCls != 0) /* The exception class was found*/
				env->ThrowNew(connectorExceptionCls, "thrown from C code");
	}
	else
	{
		// set the java member to the value we just obtained
		jclass objectCls = env->FindClass("org/n52/connectors/impl/GdalConnector");
		jfieldID fid = env->GetFieldID(objectCls, "gdalDataSet", "J"); // "J" is type "long"
		env->SetLongField(obj, fid, (jlong)gdalDataSet);
	}
}

/*
 * Class:     org_n52_connectors_impl_GdalConnector
 * Method:    createGdal
 * Signature: (Ljava/lang/String;Ljava/lang/String;III)V
 */
JNIEXPORT void JNICALL Java_org_n52_connectors_impl_GdalConnector_createGdal
  (JNIEnv * env, jobject obj,
		jstring javaFileName, jstring javaDriverName, jstring javaDataType,
		jint nPixels, jint nLines, jint nBands,
		jstring javaSrsWkt, jdoubleArray javaGeoTransform)
{
	if (!gdalInitialized)
	{
		GDALAllRegister();
		gdalInitialized = true;
	}

	const char *strFileName = env->GetStringUTFChars(javaFileName, 0);
	const char *strDriver = env->GetStringUTFChars(javaDriverName, 0);
	const char *strDataType = env->GetStringUTFChars(javaDataType, 0);
	const char *strSrsWkt = env->GetStringUTFChars(javaSrsWkt, 0);

	GDALDriver *poDriver = GetGDALDriverManager()->GetDriverByName(strDriver);
	/*
  if( poDriver == NULL )
      poDriver = GetGDALDriverManager()->GetDriverByName( "GTiff" );
	*/

	GDALDataType gdt = GDT_Unknown;
	if (stricmp("byte", strDataType) == 0)
		gdt = GDT_Byte; /*! Eight bit unsigned integer */
	else if (stricmp("uint16", strDataType) == 0)
		gdt = GDT_UInt16; /*! Sixteen bit unsigned integer */
	else if (stricmp("int16", strDataType) == 0)
		gdt = GDT_Int16; /*! Sixteen bit signed integer */
	else if (stricmp("uint32", strDataType) == 0)
		gdt = GDT_UInt32; /*! Thirty two bit unsigned integer */
	else if (stricmp("int32", strDataType) == 0)
		gdt = GDT_Int32; /*! Thirty two bit signed integer */
	else if (stricmp("float32", strDataType) == 0)
		gdt = GDT_Float32; /*! Thirty two bit floating point */
	else if (stricmp("float64", strDataType) == 0)
		gdt = GDT_Float64; /*! Sixty four bit floating point */
	else if (stricmp("cint16", strDataType) == 0)
		gdt = GDT_CInt16; /*! Complex Int16 */
	else if (stricmp("cint32", strDataType) == 0)
		gdt = GDT_CInt32; /*! Complex Int32 */
	else if (stricmp("cfloat32", strDataType) == 0)
		gdt = GDT_CFloat32; /*! Complex Float32 */
	else if (stricmp("cfloat64", strDataType) == 0)
		gdt = GDT_CFloat64; /*! Complex Float64 */

	char **papszCreateOptions = NULL;
	GDALDataset* gdalDataSet = poDriver->Create(strFileName, nPixels, nLines, nBands, gdt, papszCreateOptions);

	if (0 == gdalDataSet)
	{
			jclass connectorExceptionCls = env->FindClass("org/n52/connectors/core/exceptions/IllegalConnectorRequest");
			if (connectorExceptionCls != 0) /* The exception class was found*/
				env->ThrowNew(connectorExceptionCls, "thrown from C code");
	}
	else
	{
		// set the java member to the value we just obtained
		jclass objectCls = env->FindClass("org/n52/connectors/impl/GdalConnector");
		jfieldID fid = env->GetFieldID(objectCls, "gdalDataSet", "J"); // "J" is type "long"
		env->SetLongField(obj, fid, (jlong)gdalDataSet);
	}

	if (env->GetArrayLength(javaGeoTransform) == 6)
	{
		jdouble * adfGeoTransform = env->GetDoubleArrayElements(javaGeoTransform, 0);
		gdalDataSet->SetGeoTransform( adfGeoTransform );
		env->ReleaseDoubleArrayElements(javaGeoTransform, adfGeoTransform, 0);
	}

	if (strlen(strSrsWkt) > 0)
		gdalDataSet->SetProjection( strSrsWkt );

	env->ReleaseStringUTFChars(javaSrsWkt, strSrsWkt);
	env->ReleaseStringUTFChars(javaDataType, strDataType);
	env->ReleaseStringUTFChars(javaDriverName, strDriver);
	env->ReleaseStringUTFChars(javaFileName, strFileName);
}

/*
 * Class:     org_n52_connectors_impl_GdalConnector
 * Method:    closeGdal
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_n52_connectors_impl_GdalConnector_closeGdal
  (JNIEnv * env, jobject obj)
{
	// obtain our GDALDataSet from the java member

	jclass objectCls = env->FindClass("org/n52/connectors/impl/GdalConnector");
	jfieldID fid = env->GetFieldID(objectCls, "gdalDataSet", "J"); // "J" is type "long"
	GDALDataset* gdalDataSet = (GDALDataset*)env->GetLongField(obj, fid);
	
	if (0 != gdalDataSet)
	{
		delete gdalDataSet;	
		env->SetLongField(obj, fid, 0L); // set to null, prevent further use
	}
}
