package org.n52.operation.gs.spatial.location.sort;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import org.geotools.feature.Feature;

import com.vividsolutions.jts.algorithm.CentroidArea;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.CoordinateSequence;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequence;
import com.vividsolutions.jts.operation.distance.DistanceOp;
import com.vividsolutions.jts.operation.distance.GeometryLocation;

public class AdjacencyListByGeometry implements AdjacencyList {

	public enum Methods {
		byAnchor, byNeighbourhood
	};

	Methods method = AdjacencyListByGeometry.Methods.byAnchor;

	Geometry anchor = null;

	String attribute;

	HashMap<String, List<Integer>> neighbours;

	List<Feature> adjacencyList;

	public class FeatureDistance implements Comparable<FeatureDistance> {

		public Feature feature;

		public Coordinate index;

		public Coordinate centroid;

		public int compareTo(FeatureDistance fd) {
			if (index.x < fd.index.x) {
				return -1;
			}
			if (index.x > fd.index.x) {
				return 1;
			}
			return 0;
		}
	};

	public class FeatureByAttribute implements Comparable<FeatureByAttribute> {

		public Feature feature;

		public String attribute;

		public int compareTo(FeatureByAttribute fd) {

			try {
				Comparable obj1 = (Comparable) feature.getAttribute(attribute);
				Comparable obj2 = (Comparable) fd.feature
						.getAttribute(attribute);
				return obj1.compareTo(obj2);
			} catch (ClassCastException ex) {
				return 0;
			}
		}

	};

	public List<Feature> getAdjecencyList(List<Feature> inputList) {

		if (method == AdjacencyListByGeometry.Methods.byAnchor)
			return sortByAnchor(inputList);
		if (method == AdjacencyListByGeometry.Methods.byNeighbourhood)
			return sortByNeighbourhood(inputList);

		return null;
	}

	private List<Feature> sortByAnchor(List<Feature> features) {
		List<Feature> sorted = new ArrayList<Feature>();
		List<FeatureDistance> sorting = new ArrayList<FeatureDistance>();
		List<FeatureByAttribute> fbaList = new ArrayList<FeatureByAttribute>();

		GeometryFactory geofac = features.get(0).getDefaultGeometry()
				.getFactory();

		for (Feature f : features) {
			FeatureByAttribute fba = new FeatureByAttribute();
			fba.feature = f;
			fba.attribute = attribute;
			fbaList.add(fba);
		}
		Collections.sort(fbaList);
		int size = fbaList.size();
		int index = 0;
		Object attribOld = null;
		CentroidArea ca = new CentroidArea();

		while (index < size) {
			// calc the centroid for each geometry
			FeatureByAttribute fba = fbaList.get(index);
			Object attrib = fba.feature.getAttribute(attribute);
			if (attrib.equals(attribOld) || attribOld == null) {
				ca.add(fba.feature.getDefaultGeometry());
				++index;
			} else {

				Coordinate crd = ca.getCentroid();
				// turn the centroid into a Point
				Coordinate[] coords = new Coordinate[1];
				coords[0] = crd;
				CoordinateSequence seq = new CoordinateArraySequence(coords);
				Point pnt = new Point(seq, geofac);

				// calculate the distance of the centroid to the line
				DistanceOp dop = new DistanceOp(pnt, anchor);
				// get the point on the line closest to the centroid
				GeometryLocation loc = dop.closestLocations()[1];
				Coordinate cross = loc.getCoordinate();

				// fill the FeatureDistance array
				FeatureDistance fd = new FeatureDistance();
				fd.feature = fba.feature;
				fd.centroid = crd;
				fd.index = cross;
				sorting.add(fd);
				ca = new CentroidArea();
			}
			attribOld = attrib;
		}
		Collections.sort(sorting);

		for (FeatureDistance fd : sorting) {
			sorted.add(fd.feature);
		}

		return sorted;
	}

	private List<Feature> sortByNeighbourhood(List<Feature> features) {
		List<Feature> sortedFeatures = sortByAnchor(features);
		Feature startFeature = sortedFeatures.get(0);
		neighbours = new HashMap<String, List<Integer>>();

		for (Feature feature : sortedFeatures) {
			for (Feature candidate : features) {
				if (feature.getID().equals(candidate.getID()))
					continue;
				Geometry currentGeometry = feature.getDefaultGeometry();
				if (currentGeometry.touches(candidate.getDefaultGeometry())) {
					if (neighbours.get(feature.getID()) == null) {
						neighbours.put(feature.getID(),
								new ArrayList<Integer>());
					}

					if (!neighbours.get(feature.getID()).contains(candidate)) {
						neighbours.get(feature.getID()).add(
								sortedFeatures.indexOf(candidate));
					}

				}
			}
		}
		Set<String> keys = neighbours.keySet();
		for (String key : keys) {
			List<Integer> list = neighbours.get(key);
			Collections.sort(list);
			neighbours.put(key, list);
		}

		adjacencyList = new ArrayList<Feature>();
		Feature currentFeature = startFeature;
		// add the first one; we have to test adjecency against something
		adjacencyList.add(startFeature);
		ArrayList<Feature> available = new ArrayList();
		available.addAll(sortedFeatures);
		while (available.size() > 1) {
			moveToNext(sortedFeatures, available, currentFeature);
			// the first element after the start element (0), which is the start
			// element, is the first one that is still available to be included.
			// All elements that where in the beginning in front of it have been
			// removed because they where in adajacency
			if (available.size() > 1)
				currentFeature = available.get(1);

		}

		return adjacencyList;

	}

	private void moveToNext(final List<Feature> sortedFeatures,
			List<Feature> available, Feature currentFeature) {
		List<Integer> neighbourList = neighbours.get(currentFeature.getID());
		if (neighbourList != null) {
			for (Integer index : neighbourList) {
				Feature candidate = sortedFeatures.get(index);
				if (!adjacencyList.contains(candidate)) {
					adjacencyList.add(candidate);
					available.remove(candidate);
					moveToNext(sortedFeatures, available, candidate);
				}
			}
		} else // this was an isolated island; add it and remove it from the
				// available list
		{
			if (!adjacencyList.contains(currentFeature)) {
				adjacencyList.add(currentFeature);
				available.remove(currentFeature);
			}
		}
	}

	public void setAnchor(Geometry anchor) {
		this.anchor = anchor;
	}

	public void setMethod(Methods method) {
		this.method = method;
	}

	public void setAttribute(String attribute) {
		this.attribute = attribute;
	}

}
